// noinspection JSCheckFunctionSignatures

import { writable } from 'svelte/store'
import { backendAddressAndPort, queryMode } from "./constants.js";

const mockItems = [
  {
    "eventType": "Television",
    "broughtToYouBy": "AMC",
    "title": "Breaking Bad"
  },
  {
    "eventType": "Television",
    "broughtToYouBy": "CBS",
    "title": "Breaking Bad Bang Theory"
  },
  {
    "eventType": "Movie",
    "broughtToYouBy": "20th Century Fox",
    "title": "Star Wars 2: The Star Warsening, brought to you by the Big Stink Corporation, a division of PepsiCo and somehow also Disney."
  },
].map(o => ({ value: o, timestamp: new Date().getTime() }))

export const testMode = true
export const state = writable({
  items: [],
  itemCount: undefined,
  error: 'Connecting to WebSocket...',
})
const updateClearError = updater => state.update(s => {
  s.error = null
  return updater(s)
})

let itemLimit = Infinity
let ws;
let disconnected = false

const getRandomFromArray = array => array[Math.floor(Math.random() * array.length)]

const testTimeout = 200
const testQuery = (mode, jsFilter, queryCode) => {
  try {
    const f = new Function('message', 'value', queryCode)
    if (mode === queryMode.REAL_TIME) {
      const addItem = () => {
        const item = getRandomFromArray(mockItems)
        if (!jsFilter || f(item, item.value)) {
          item.timestamp = new Date().getTime()
          state.update(s => ({ ...s, items: [item, ...s.items].slice(0, itemLimit), itemCount: 0 }))
        }
        setTimeout(addItem, testTimeout)
      }
      setTimeout(addItem, testTimeout)
    } else {
      state.update(s => ({ ...s, items: mockItems.filter(item => !jsFilter || f(item, item.value)), itemCount: 0 }))
    }
  } catch (e) {
    console.log('Caught an error:', e.toString())
    state.update(s => ({ ...s, error: e.toString() }))
  }
}

// noinspection JSUnusedGlobalSymbols
export const query = async ({ cluster, topic, mode, jsFilter, queryCode, maxItems, mutableObjects }) => {
  updateClearError(s => ({ ...s, items: [], itemCount: 0 }))
  if (testMode) {
    testQuery(mode, jsFilter, queryCode)
    return
  }
  if (disconnected) {
    connect()
    disconnected = false
  }
  updateClearError(s => ({ ...s, items: [], itemCount: 0 }))
  itemLimit = maxItems
  ws.send(JSON.stringify({
    cluster,
    mode,
    topic,
    maxItems,
    immutable: !mutableObjects,
    searchCode: jsFilter && queryCode
  }))
}

export const connect = () => {
  try {
    ws = new WebSocket(`ws://${backendAddressAndPort}`)
  } catch (e) {
    console.error(e.toString())
  }
  if (!ws) {
    updateClearError(s => ({ ...s, error: 'Unable to connect to websocket.' }))
    return
  }

  ws.addEventListener('open', () => {
    console.log('WebSocket opened')
  })

  ws.addEventListener('message', message => {
    let data;
    try {
      data = JSON.parse(message.data)
      switch (data?.type.toLowerCase()) {
        case 'complete':
          updateClearError(s => ({
            ...s,
            items: data.message
          }))
          break;
        case 'item_count':
          updateClearError(s => ({
            ...s,
            itemCount: console.log('state', s) || data.message
          }))
          break;
        case 'message':
          updateClearError(s => ({
            ...s,
            items: console.log('new item', data.message) || [data.message, ...s.items].slice(0, itemLimit)
          }))
          break;
      }
    } catch (e) {
      data = message.data
    }
  })

  ws.addEventListener('close', () => {
    disconnected = true
    state.update(s => ({ ...s, error: 'WebSocket connection closed.' }))
  })

  ws.addEventListener('error', () => {
    state.update(s => ({ ...s, error: 'WebSocket connection error.' }))
  })

  if (ws.readyState === ws.CLOSED || ws.readyState === ws.CLOSING) {
    state.update(s => ({ ...s, error: 'WebSocket connection closed.' }))
  }
}
